/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aacdec.c paraslash's aac (m4a) decoder */

#include "para.h"

#include "list.h"
#include "sched.h"
#include "filter.h"
#include "error.h"
#include "string.h"
#include "aac.h"

/** the output buffer size */
#define AAC_OUTBUF_SIZE (32 * 1024)

#define MAX_ERRORS 20

/**
 * data specific to the aacdec filter
 *
 * \sa filter, filter_node
 */
struct private_aacdec_data {
	NeAACDecHandle handle;
	NeAACDecFrameInfo frame_info;

	int initialized;
	int decoder_length;
	unsigned error_count;
	size_t consumed_total;
	size_t entry;
};

static ssize_t aacdec(char *input_buffer, size_t len, struct filter_node *fn)
{
	struct private_aacdec_data *padd = fn->private_data;
	struct filter_chain *fc = fn->fc;
	int i, ret;
	unsigned char *p, *outbuffer;
	unsigned char *inbuf = (unsigned char*)input_buffer;
	size_t skip, consumed = 0;

	if (fn->loaded > fn->bufsize * 3 / 5)
		return 0;
	if (len < 2048 && !*fc->input_eof)
		return 0;

	if (!padd->initialized) {
		unsigned long rate = 0;
		unsigned char channels = 0;
		padd->decoder_length = aac_find_esds(inbuf, len, &skip);
		PARA_INFO_LOG("decoder len: %d\n", padd->decoder_length);
		if (padd->decoder_length < 0) {
			ret = NeAACDecInit(padd->handle, inbuf,
				len, &rate, &channels);
			PARA_INFO_LOG("decoder init: %d\n", ret);
			if (ret < 0) {
				ret = -E_AACDEC_INIT;
				goto out;
			}
			consumed = ret;
		} else {
			consumed += skip;
			p = inbuf + consumed;
			ret = -E_AACDEC_INIT;
			if (NeAACDecInit2(padd->handle, p,
					padd->decoder_length, &rate,
					&channels) < 0)
				goto out;
		}
		fc->samplerate = rate;
		fc->channels = channels;
		PARA_INFO_LOG("rate: %u, channels: %d\n",
			fc->samplerate, fc->channels);
		padd->initialized = 1;
	}
	if (padd->decoder_length > 0) {
		consumed = 0;
		if (!padd->entry) {
			ret = aac_find_entry_point(inbuf + consumed,
				len - consumed, &skip);
			if (ret < 0) {
				ret = len;
				goto out;
			}
			consumed += skip;
			padd->entry = ret;
			PARA_INFO_LOG("entry: %zu\n", padd->entry);
		}
		ret = len;
		if (padd->consumed_total + len < padd->entry)
			goto out;
		if (padd->consumed_total < padd->entry)
			consumed = padd->entry - padd->consumed_total;
	}
	for (; consumed < len; consumed++)
		if ((inbuf[consumed] & 0xfe) == 0x20)
			break;
	if (consumed >= len)
		goto success;
	p = inbuf + consumed;
	outbuffer = NeAACDecDecode(padd->handle, &padd->frame_info, p,
		len - consumed);
	if (padd->frame_info.error) {
		ret = -E_AAC_DECODE;
		if (padd->error_count++ > MAX_ERRORS)
			goto out;
		PARA_ERROR_LOG("frame_error: %d, consumed: %zu + %zd + %lu\n",
			padd->frame_info.error, padd->consumed_total,
			consumed, padd->frame_info.bytesconsumed);
		PARA_ERROR_LOG("%s\n", NeAACDecGetErrorMessage(
			padd->frame_info.error));
		consumed++; /* catch 21 */
		goto success;
	}
	padd->error_count = 0;
	consumed += padd->frame_info.bytesconsumed;
	ret = consumed;
	if (!padd->frame_info.samples)
		goto out;
	ret = -E_AAC_OVERRUN;
	if (padd->frame_info.samples * 2 + fn->loaded > fn->bufsize)
		goto out;
	for (i = 0; i < padd->frame_info.samples; i++) {
		short *s = (short *)outbuffer;
		fn->buf[fn->loaded++] = s[i] & 0xff;
		fn->buf[fn->loaded++] = (s[i] >> 8) & 0xff;
	}
success:
	ret = consumed;
out:
	if (ret > 0)
		padd->consumed_total += ret;
	return ret;
}

static void aacdec_open(struct filter_node *fn)
{
	fn->private_data = para_calloc(sizeof(struct private_aacdec_data));
	struct private_aacdec_data *padd = fn->private_data;

	fn->bufsize = AAC_OUTBUF_SIZE;
	fn->buf = para_calloc(fn->bufsize);
	padd->handle = aac_open();
}

static void aacdec_close(struct filter_node *fn)
{
	struct private_aacdec_data *padd = fn->private_data;

	NeAACDecClose(padd->handle);
	free(fn->buf);
	fn->buf = NULL;
	free(padd);
	fn->private_data = NULL;
}

/**
 * the init function of the aacdec filter
 *
 * \sa filter::init
 */
void aacdec_init(struct filter *f)
{
	f->open = aacdec_open;
	f->convert = aacdec;
	f->close = aacdec_close;
}
